__all__ = [
    "streamline_event_times",
    "streamline_event",
    "ordered_insert",
    "TimelineOMat",
    "SkipEvent",
    "SkipInvalidEvent",
    "SkipOccludedEvent",
    "NoCallAllowedError",
    "PositionsOffsetsTuple",
    "TimeRangeTuple",
]

from collections.abc import Callable, Iterable, Iterator, MutableSequence, Sequence
from datetime import datetime as dt
from datetime import timezone as tz
from functools import lru_cache
from itertools import chain, tee
from typing import Literal, NamedTuple, NewType, TypedDict, TypeVar, Unpack, cast


class TimeRangeTuple(NamedTuple):
    start: dt
    stop: dt


Event = TypeVar("Event")
Offset = NewType("Offset", int)
Position = NewType("Position", int)
ExtractionResult = dt | float | int | str
FilterFunction = Callable[[Event], bool]
CutHandler = Callable[[TimeRangeTuple, TimeRangeTuple, Iterator[Event]], TimeRangeTuple]
CallableExtractor = Callable[[Event], ExtractionResult]
Extractor = str | CallableExtractor
CallableSetter = Callable[[Event, dt], None]
Setter = str | CallableSetter


class PositionsOffsetsTuple(NamedTuple):
    positions: Sequence[Position]
    offsets: Sequence[Offset]


class SkipEvent(BaseException):
    pass


class SkipInvalidEvent(SkipEvent):
    pass


class SkipEmptyEvent(SkipEvent):
    pass


class SkipOccludedEvent(SkipEvent):
    original: TimeRangeTuple

    def __init__(self, *args, original: TimeRangeTuple, **kwargs):
        super().__init__(*args, **kwargs)
        self.original = original


class NoCallAllowedError(Exception):
    pass


# old name
NewTimesResult = TimeRangeTuple


def default_cut_handler(new: TimeRangeTuple, orig: TimeRangeTuple, remaining: Iterator[Event]) -> TimeRangeTuple:
    raise SkipEvent("Noop - don't handle cut")


def create_extractor(extractor: Extractor) -> CallableExtractor:
    if not isinstance(extractor, str):
        return extractor

    def _extractor(event: Event) -> ExtractionResult:
        try:
            if isinstance(event, dict):
                return event[extractor]
            return getattr(event, extractor)
        except (KeyError, AttributeError) as exc:
            raise SkipInvalidEvent from exc

    return _extractor


# for allowing chaining in higher code levels, expose disallow_*
def create_setter(
    setter: Setter,
    *,
    disallow_call: bool = False,
    disallow_call_instant: bool = False,
) -> CallableSetter:
    if not isinstance(setter, str):
        if disallow_call_instant:
            raise NoCallAllowedError("extractor is not a string and no setter is set")
        if disallow_call:

            def _setter(event: Event, value: dt) -> None:  # noqa: RET505
                raise NoCallAllowedError("extractor is not a string and no setter is set")

            return _setter
        return setter

    # setter is string
    def _setter_str(event: Event, value: dt) -> None:
        if isinstance(event, dict):
            event[setter] = value
        else:
            setattr(event, setter, value)

    return _setter_str


@lru_cache(1024, typed=True)
def handle_result(result: ExtractionResult, fallback_timezone: tz | None = None) -> dt:
    if isinstance(result, dt):
        if fallback_timezone and not result.tzinfo:
            result = result.replace(tzinfo=fallback_timezone)
        return result
    elif isinstance(result, int | float):
        return dt.fromtimestamp(result, fallback_timezone)
    elif isinstance(result, str):  # type: ignore
        return handle_result(dt.fromisoformat(result), fallback_timezone=fallback_timezone)
    else:
        raise TypeError(f"not supported type: {type(result)}")


def extract_tuple_from_event(
    event: Event,
    *,
    start_extractor: CallableExtractor,
    stop_extractor: CallableExtractor,
    fallback_timezone: tz | None = None,
    ensure_timespan: bool = False,
) -> TimeRangeTuple:
    start = handle_result(start_extractor(event), fallback_timezone=fallback_timezone)
    stop = handle_result(stop_extractor(event), fallback_timezone=fallback_timezone)

    if stop < start:
        raise SkipInvalidEvent("duration of the event is < 0")
    if ensure_timespan and stop == start:
        raise SkipEmptyEvent("duration of the event is = 0")

    return TimeRangeTuple(start=start, stop=stop)


def _array_window(array: Sequence[Event], offset: Offset, direction: Literal["asc", "desc"]):
    length = len(array)
    if direction == "asc":
        for pos in range(offset, length):
            yield array[pos]
    else:
        for pos in range(offset, length):
            yield array[length - pos - 1]


def _streamline_event_times(
    event: Event,
    timeline: Iterator[Event] | None,
    cut_handler: CutHandler,
    start_extractor: Extractor = "start",
    stop_extractor: Extractor = "stop",
    ensure_timespan: bool = False,
    filter_fn: FilterFunction | None = None,
    fallback_timezone: tz | None = None,
) -> tuple[TimeRangeTuple, TimeRangeTuple]:
    start_extractor = create_extractor(start_extractor)
    stop_extractor = create_extractor(stop_extractor)
    start, stop = orig_tuple = extract_tuple_from_event(
        event,
        start_extractor=start_extractor,
        stop_extractor=stop_extractor,
        fallback_timezone=fallback_timezone,
        ensure_timespan=ensure_timespan,
    )
    if not timeline:
        return orig_tuple, orig_tuple
    is_initial_empty = start == stop
    while True:
        try:
            ev = next(timeline)
        except StopIteration:
            break
        # only check events which are not filtered
        if filter_fn and not filter_fn(ev):
            continue
        # get event time tuple
        try:
            ev_start, ev_stop = extract_tuple_from_event(
                ev,
                start_extractor=start_extractor,
                stop_extractor=stop_extractor,
                fallback_timezone=fallback_timezone,
                ensure_timespan=False,
            )
        except SkipEvent:
            continue
        # current event is within an existing event
        if ev_start <= start and ev_stop >= stop:
            raise SkipOccludedEvent(original=orig_tuple)
        # existing event is within this event, we need to cut or raise
        if start < ev_start and stop > ev_stop:
            assert start < stop
            # split iterator, if timeline is a tee object, there is already an optimization
            timeline, remaining = tee(timeline)
            start, stop = cut_handler(TimeRangeTuple(start, stop), orig_tuple, chain([ev], remaining))
        # current event is at the end of an existing event overlapping
        if ev_start <= start and ev_stop > start:
            start = ev_stop
        # current event is at the start of an existing event overlapping
        if ev_start < stop and ev_stop >= stop:
            stop = ev_start
        # check if there is still a valid timespan left after if not initial empty
        if not is_initial_empty and stop <= start:
            raise SkipOccludedEvent(original=orig_tuple)
    return TimeRangeTuple(start=start, stop=stop), orig_tuple


class _streamline_event_base_kwargs(TypedDict, total=False):
    filter_fn: FilterFunction | None
    fallback_timezone: tz | None


class _streamline_event_times_kwargs(_streamline_event_base_kwargs):
    occlusions: list[TimeRangeTuple] | None
    start_extractor: Extractor
    stop_extractor: Extractor
    ensure_timespan: bool
    cut_handler: CutHandler


def streamline_event_times(
    event: Event,
    *timelines: Iterable[Event],
    occlusions: list[TimeRangeTuple] | None = None,
    start_extractor: Extractor = "start",
    stop_extractor: Extractor = "stop",
    ensure_timespan: bool = False,
    cut_handler: CutHandler = default_cut_handler,
    **kwargs: Unpack[_streamline_event_base_kwargs],
) -> TimeRangeTuple:
    try:
        new_tuple, orig_tuple = _streamline_event_times(
            event,
            # None will trigger a shortcut
            chain.from_iterable(timelines) if timelines else None,
            start_extractor=start_extractor,
            stop_extractor=stop_extractor,
            ensure_timespan=ensure_timespan,
            cut_handler=cut_handler,
            **kwargs,
        )
    except SkipOccludedEvent as exc:
        if occlusions is not None:
            occlusions.append(exc.original)
        raise exc
    if new_tuple != orig_tuple and occlusions is not None:
        if orig_tuple.start != new_tuple.start:
            occlusions.append(TimeRangeTuple(start=orig_tuple.start, stop=new_tuple.start))
        if orig_tuple.stop != new_tuple.stop:
            occlusions.append(TimeRangeTuple(start=new_tuple.stop, stop=orig_tuple.stop))
    return new_tuple


class _streamline_event_kwargs(_streamline_event_base_kwargs):
    start_extractor: Extractor
    stop_extractor: Extractor
    start_setter: Setter | None
    stop_setter: Setter | None
    occlusions: list[TimeRangeTuple] | None
    ensure_timespan: bool
    cut_handler: CutHandler


def streamline_event(
    event: Event,
    *timelines: Iterable[Event],
    start_extractor: Extractor = "start",
    stop_extractor: Extractor = "stop",
    start_setter: Setter | None = None,
    stop_setter: Setter | None = None,
    occlusions: list[TimeRangeTuple] | None = None,
    ensure_timespan: bool = False,
    cut_handler: CutHandler = default_cut_handler,
    **kwargs: Unpack[_streamline_event_base_kwargs],
) -> Event:
    if not timelines:
        return event
    if start_setter is not None:
        start_setter = create_setter(start_setter)
    else:
        # because of disallow_call_instant we correctly raise for non-strings
        start_setter = create_setter(cast(str, start_extractor), disallow_call_instant=True)
    if stop_setter is not None:
        stop_setter = create_setter(stop_setter)
    else:
        # because of disallow_call_instant we correctly raise for non-strings
        stop_setter = create_setter(cast(str, stop_extractor), disallow_call_instant=True)
    new_tuple = streamline_event_times(
        event,
        *timelines,
        start_extractor=start_extractor,
        stop_extractor=stop_extractor,
        occlusions=occlusions,
        ensure_timespan=ensure_timespan,
        cut_handler=cut_handler,
        **kwargs,
    )
    start_setter(event, new_tuple.start)
    stop_setter(event, new_tuple.stop)
    return event


class _transform_events_to_times_kwargs(_streamline_event_base_kwargs):
    start_extractor: Extractor
    stop_extractor: Extractor


def transform_events_to_times(
    *timelines: Iterable[Event],
    start_extractor: Extractor = "start",
    stop_extractor: Extractor = "stop",
    filter_fn: FilterFunction | None = None,
    fallback_timezone: tz | None = None,
) -> Iterable[tuple[TimeRangeTuple, Event]]:
    start_extractor = create_extractor(start_extractor)
    stop_extractor = create_extractor(stop_extractor)
    if not timelines:
        return
    for ev in chain.from_iterable(timelines):
        if filter_fn and not filter_fn(ev):
            continue
        try:
            retval = (
                extract_tuple_from_event(
                    ev,
                    start_extractor=start_extractor,
                    stop_extractor=stop_extractor,
                    fallback_timezone=fallback_timezone,
                    ensure_timespan=False,
                ),
                ev,
            )
            yield retval
        except SkipEvent:
            continue


def _ordered_insert(
    event: Event,
    event_times: TimeRangeTuple,
    timeline: MutableSequence[Event],
    offset: Offset,
    direction: Literal["asc", "desc"],
    start_extractor: CallableExtractor,
    stop_extractor: CallableExtractor,
    fallback_timezone: tz | None,
    no_insert: bool,
) -> tuple[bool, Position]:
    if not len(timeline):
        if not no_insert:
            timeline.append(event)
        return no_insert, cast(Position, 0)
    last_pos = None
    length = len(timeline)
    for position in range(offset, length):
        if direction == "desc":
            position = length - position - 1
        if last_pos is None:
            last_pos = position
        ev = timeline[position]
        try:
            ev_times = extract_tuple_from_event(
                ev,
                start_extractor=start_extractor,
                stop_extractor=stop_extractor,
                fallback_timezone=fallback_timezone,
                # consider empty events
                ensure_timespan=False,
            )
        except SkipEvent:
            last_pos = position
            continue
        if direction == "asc":
            if ev_times > event_times:
                if not no_insert:
                    timeline.insert(position, event)
                return no_insert, cast(Position, position)
        else:
            if (ev_times[1], ev_times[0]) < (event_times[1], ev_times[0]):
                if not no_insert:
                    timeline.insert(cast(int, last_pos), event)
                return no_insert, cast(Position, last_pos)
        last_pos = position
    if direction == "asc":
        if not no_insert:
            timeline.append(event)
        return no_insert, cast(Position, length)
    else:
        if not no_insert:
            timeline.insert(0, event)
        return no_insert, cast(Position, 0)


class _ordered_insert_kwargs(TypedDict, total=False):
    offsets: Sequence[Offset | None | Literal[0]] | None
    no_update_timelines: set[int] | None
    start_extractor: Extractor
    stop_extractor: Extractor
    fallback_timezone: tz | None
    direction: Literal["asc", "desc"]
    ensure_timespan: bool


def ordered_insert(
    event: Event,
    *timelines: Sequence[Event],
    offsets: Sequence[Offset | None | Literal[0]] | None = None,
    no_update_timelines: set[int] | None = None,
    start_extractor: Extractor = "start",
    stop_extractor: Extractor = "stop",
    fallback_timezone: tz | None = None,
    direction: Literal["asc", "desc"] = "asc",
    ensure_timespan: bool = False,
) -> PositionsOffsetsTuple:
    start_extractor = create_extractor(start_extractor)
    stop_extractor = create_extractor(stop_extractor)
    return_positions: list[Position] = []
    return_offsets: list[Offset] = []
    event_times = extract_tuple_from_event(
        event,
        start_extractor=start_extractor,
        stop_extractor=stop_extractor,
        fallback_timezone=fallback_timezone,
        ensure_timespan=ensure_timespan,
    )
    for count, timeline in enumerate(timelines):
        no_insert, position = _ordered_insert(
            event,
            event_times=event_times,
            timeline=cast(
                MutableSequence[Event],
                timeline,
            ),
            offset=cast(Offset, 0) if offsets is None else (offsets[count] or cast(Offset, 0)),
            direction=direction,
            fallback_timezone=fallback_timezone,
            no_insert=False if no_update_timelines is None else count in no_update_timelines,
            start_extractor=start_extractor,
            stop_extractor=stop_extractor,
        )
        return_positions.append(cast(Position, position))
        if direction == "desc":
            return_offsets.append(cast(Offset, len(timeline) - position - (0 if no_insert else 1)))
        else:
            return_offsets.append(cast(Offset, position))
    return PositionsOffsetsTuple(return_positions, return_offsets)


class _streamline_ordered_insert_kwargs(_streamline_event_kwargs):
    offsets: Sequence[Offset | None | Literal[0]] | None
    no_update_timelines: set[int] | None
    direction: Literal["asc", "desc"]
    cut_direction: Literal["asc", "desc", "occlusion"] | None


def streamlined_ordered_insert(
    event: Event,
    *timelines: Sequence[Event],
    filter_fn: FilterFunction | None = None,
    offsets: Sequence[Offset | None | Literal[0]] | None = None,
    no_update_timelines: set[int] | None = None,
    direction: Literal["asc", "desc"] = "asc",
    cut_handler: CutHandler = default_cut_handler,
    start_extractor: Extractor = "start",
    stop_extractor: Extractor = "stop",
    start_setter: Setter | None = None,
    stop_setter: Setter | None = None,
    fallback_timezone: tz | None = None,
    occlusions: list[TimeRangeTuple] | None = None,
    ensure_timespan: bool = False,
) -> PositionsOffsetsTuple:
    if start_setter is not None:
        start_setter = create_setter(start_setter)
    else:
        # because of disallow_call_instant we correctly raise for non-strings
        start_setter = create_setter(cast(str, start_extractor), disallow_call_instant=True)
    if stop_setter is not None:
        stop_setter = create_setter(stop_setter)
    else:
        # because of disallow_call_instant we correctly raise for non-strings
        stop_setter = create_setter(cast(str, stop_extractor), disallow_call_instant=True)
    # must be after setter extractors
    start_extractor = create_extractor(start_extractor)
    stop_extractor = create_extractor(stop_extractor)
    return ordered_insert(
        streamline_event(
            event,
            *(
                _array_window(
                    timeline, cast(Offset, 0) if offsets is None else (offsets[count] or cast(Offset, 0)), direction
                )
                for count, timeline in enumerate(timelines)
            ),
            start_extractor=start_extractor,
            stop_extractor=stop_extractor,
            start_setter=start_setter,
            stop_setter=stop_setter,
            filter_fn=filter_fn,
            occlusions=occlusions,
            fallback_timezone=fallback_timezone,
            ensure_timespan=ensure_timespan,
            cut_handler=cut_handler,
        ),
        *timelines,
        no_update_timelines=no_update_timelines,
        start_extractor=start_extractor,
        stop_extractor=stop_extractor,
        offsets=offsets,
        direction=direction,
        fallback_timezone=fallback_timezone,
    )


class TimelineOMat:
    start_extractor: CallableExtractor
    stop_extractor: CallableExtractor
    start_setter: CallableSetter
    stop_setter: CallableSetter
    no_update_timelines: set[int] | None
    filter_fn: FilterFunction | None
    fallback_timezone: tz | None
    direction: Literal["asc", "desc"]
    cut_handler: CutHandler
    ensure_timespan: bool

    def __init__(
        self,
        *,
        start_extractor: Extractor = "start",
        stop_extractor: Extractor = "stop",
        start_setter: Setter | None = None,
        stop_setter: Setter | None = None,
        no_update_timelines: set[int] | None = None,
        filter_fn: FilterFunction | None = None,
        fallback_timezone: tz | None = None,
        # for ordered_insert
        direction: Literal["asc", "desc"] = "asc",
        cut_handler: CutHandler = default_cut_handler,
        ensure_timespan: bool = False,
    ):
        self.start_extractor = create_extractor(start_extractor)
        self.stop_extractor = create_extractor(stop_extractor)
        self.no_update_timelines = no_update_timelines
        self.filter_fn = filter_fn
        self.fallback_timezone = fallback_timezone
        self.direction = direction
        self.cut_handler = cut_handler
        self.ensure_timespan = ensure_timespan
        if start_setter is not None:
            self.start_setter = create_setter(start_setter)
        else:
            # because of disallow_call_instant we correctly raise for non-strings when called
            self.start_setter = create_setter(cast(str, start_extractor), disallow_call=True)
        if stop_setter is not None:
            self.stop_setter = create_setter(stop_setter)
        else:
            # because of disallow_call_instant we correctly raise for non-strings when called
            self.stop_setter = create_setter(cast(str, stop_extractor), disallow_call=True)

    def streamline_event_times(
        self, event: Event, *timelines: Iterable[Event], **kwargs: Unpack[_streamline_event_times_kwargs]
    ) -> TimeRangeTuple:
        return streamline_event_times(
            event,
            *timelines,
            occlusions=kwargs.get("occlusions"),
            cut_handler=kwargs.get("cut_handler", self.cut_handler),
            ensure_timespan=kwargs.get("ensure_timespan", self.ensure_timespan),
            start_extractor=kwargs.get("start_extractor", self.start_extractor),
            stop_extractor=kwargs.get("stop_extractor", self.stop_extractor),
            filter_fn=kwargs.get("filter_fn", self.filter_fn),
            fallback_timezone=kwargs.get("fallback_timezone", self.fallback_timezone),
        )

    def streamline_event(
        self, event: Event, *timelines: Iterable[Event], **kwargs: Unpack[_streamline_event_kwargs]
    ) -> Event:
        if not timelines:
            return event
        return streamline_event(
            event,
            *timelines,
            occlusions=kwargs.get("occlusions"),
            cut_handler=kwargs.get("cut_handler", self.cut_handler),
            ensure_timespan=kwargs.get("ensure_timespan", self.ensure_timespan),
            start_extractor=kwargs.get("start_extractor", self.start_extractor),
            stop_extractor=kwargs.get("stop_extractor", self.stop_extractor),
            filter_fn=kwargs.get("filter_fn", self.filter_fn),
            fallback_timezone=kwargs.get("fallback_timezone", self.fallback_timezone),
            start_setter=kwargs.get("start_setter", self.start_setter),
            stop_setter=kwargs.get("stop_setter", self.stop_setter),
        )

    def transform_events_to_times(
        self, *timelines: Iterable[Event], **kwargs: Unpack[_transform_events_to_times_kwargs]
    ) -> Iterable[tuple[TimeRangeTuple, Event]]:
        if not timelines:
            return []
        return transform_events_to_times(
            *timelines,
            start_extractor=kwargs.get("start_extractor", self.start_extractor),
            stop_extractor=kwargs.get("stop_extractor", self.stop_extractor),
            filter_fn=kwargs.get("filter_fn", self.filter_fn),
            fallback_timezone=kwargs.get("fallback_timezone", self.fallback_timezone),
        )

    def ordered_insert(
        self,
        event: Event,
        *timelines: Sequence[Event],
        **kwargs: Unpack[_ordered_insert_kwargs],
    ) -> PositionsOffsetsTuple:
        return ordered_insert(
            event,
            *timelines,
            ensure_timespan=kwargs.get("ensure_timespan", self.ensure_timespan),
            no_update_timelines=kwargs.get("no_update_timelines", self.no_update_timelines),
            offsets=kwargs.get("offsets"),
            start_extractor=kwargs.get("start_extractor", self.start_extractor),
            stop_extractor=kwargs.get("stop_extractor", self.stop_extractor),
            direction=kwargs.get("direction", self.direction),
            fallback_timezone=kwargs.get("fallback_timezone", self.fallback_timezone),
        )

    def streamlined_ordered_insert(
        self,
        event: Event,
        *timelines: Sequence[Event],
        **kwargs: Unpack[_streamline_ordered_insert_kwargs],
    ) -> PositionsOffsetsTuple:
        return streamlined_ordered_insert(
            event,
            *timelines,
            no_update_timelines=kwargs.get("no_update_timelines", self.no_update_timelines),
            offsets=kwargs.get("offsets"),
            occlusions=kwargs.get("occlusions"),
            cut_handler=kwargs.get("cut_handler", self.cut_handler),
            ensure_timespan=kwargs.get("ensure_timespan", self.ensure_timespan),
            start_extractor=kwargs.get("start_extractor", self.start_extractor),
            stop_extractor=kwargs.get("stop_extractor", self.stop_extractor),
            direction=kwargs.get("direction", self.direction),
            fallback_timezone=kwargs.get("fallback_timezone", self.fallback_timezone),
            filter_fn=kwargs.get("filter_fn", self.filter_fn),
            start_setter=kwargs.get("start_setter", self.start_setter),
            stop_setter=kwargs.get("stop_setter", self.stop_setter),
        )



import random
from datetime import datetime as dt
from datetime import timedelta as td

import edgy
import pytest
from edgy import Registry
from edgy.testing.client import DatabaseTestClient
from faker import Faker

import timelineomat

database = DatabaseTestClient("sqlite:///test_db.sqlite", drop_database=True)
models = Registry(database=edgy.Database(database, force_rollback=True))

pytestmark = pytest.mark.anyio


@pytest.fixture(autouse=True, scope="module")
async def create_test_database():
    # this creates and drops the database
    async with database:
        await models.create_all()
        yield
        await models.drop_all()


@pytest.fixture(autouse=True, scope="function")
async def rollback_transactions():
    # this rolls back
    async with models:
        yield


class Event(edgy.Model):
    start: dt = edgy.DateTimeField()
    stop: dt = edgy.DateTimeField()
    timeline: str = edgy.CharField(max_length=20)

    class Meta:
        registry = models


def _generate_time_tuple(faker, start):
    return start, start + td(hours=faker.random_int(1, 48))


async def _generate_event_series(name: str, amount=1000):
    faker = Faker()
    ts_start = faker.past_datetime(
        "-50d",
    )
    events = []
    for _i in range(amount):
        ts_stop = _generate_time_tuple(faker, ts_start)[1]
        events.append(Event(start=ts_start, stop=ts_stop, timeline=name))
        ts_start += td(hours=faker.random_int(1, 48))
    # prevent order
    random.shuffle(events)
    await Event.query.bulk_create(events)


async def test_multiline_ordering():
    tm = timelineomat.TimelineOMat()
    await _generate_event_series("tl1", 70)
    await _generate_event_series("tl1", 70)
    await _generate_event_series("tl2", 70)
    await _generate_event_series("tl2", 70)
    timeline1 = []
    timeline2 = await Event.query.filter(timeline="tl2").order_by("start").distinct()
    timeline_save = []
    timeline_delete = []
    for event in await Event.query.filter(timeline="tl1").order_by("start").distinct():
        occlusions = []
        try:
            tm.streamlined_ordered_insert(event, timeline1, timeline2, no_update_timelines={1}, occlusions=occlusions)
        except timelineomat.SkipEvent:
            timeline_delete.append(event.id)
            continue
        if occlusions:
            if len(occlusions) == 1:
                assert occlusions[0].stop <= event.start or occlusions[0].start >= event.stop
            else:
                assert occlusions[0].stop <= event.start and occlusions[1].start >= event.stop

            timeline_save.append(event)
    await Event.query.filter(id__in=timeline_delete).delete()
    if timeline_save:
        await Event.query.bulk_update(timeline_save, {"start", "stop"})
    # check ordering array
    last_stop = None
    for event in timeline1:
        if last_stop is not None:
            assert event.start >= last_stop
        assert event.start <= event.stop
        last_stop = event.stop

    last_stop = None
    # check ordering db
    for event in await Event.query.filter(timeline="tl1").order_by("start"):
        if last_stop is not None:
            assert event.start >= last_stop
        assert event.start <= event.stop
        last_stop = event.stop



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime as dt
from typing import Any

from timelineomat import BaseTMSnapshot, SnapshotType, TMField, extract_snapshot_data

database_query: Any = []


@dataclass(kw_only=True)
class ExampleSnapshot(BaseTMSnapshot):
    snapshot_for: dt
    content_specifier: str
    data: dict[str, Any] = field(default_factory=dict, init=False)
    managed: set[str] = field(default_factory=set, init=False)
    snapshot_type: SnapshotType

    snapshot_field = TMField()

    def __post_init__(self, **kwargs):
        # fix data
        for k, v in self.__dict__.items():
            if isinstance(field := type(self).__dict__.get(k), TMField) and v is not field:
                self.data[k] = field.serializer(v)

    @classmethod
    def process_snapshot_content_specifier(cls, content_specifier: str | None):
        # optional, only required when content_specifiers need a processing
        return content_specifier or cls.__name__

    @classmethod
    def get_snapshots_impl(cls, *, content_specifier, after, before, start_snapshot, extra_arg=False):
        snapshots = [start_snapshot] if start_snapshot is not None else []
        first_snapshot_data = None
        snapshots = ...
        for snap in snapshots:
            extracted, timepoint, snap_type = extract_snapshot_data(snap, fallback_tz=cls.snapshot_fallback_tz)
            if before is not None and timepoint >= before:
                break
            if after is not None and timepoint < after:
                if snap_type == SnapshotType.full:
                    first_snapshot_data = extracted
                    first_snapshot_data["snapshot_for"] = timepoint
                elif first_snapshot_data is not None:
                    first_snapshot_data.update(extracted)
                    first_snapshot_data["snapshot_for"] = timepoint
                continue
            if snap_type != SnapshotType.full and (not snapshots or first_snapshot_data is not None):
                raise ValueError("No full snapshot found")
            if not snapshots and first_snapshot_data is not None:
                if snap_type != SnapshotType.full:
                    first_snapshot_data["snapshot_type"] = SnapshotType.temporary
                    first_snapshot_data["snapshot_for"] = after or timepoint
                    assert "snapshot_type" in first_snapshot_data
                    snapshots.append(first_snapshot_data)

                first_snapshot_data = None
            # start from the last full snapshot if both are None
            if snap_type == SnapshotType.full and before is None and after is None:
                snapshots.clear()
                first_snapshot_data = None
            snapshots.append(snap)
        if first_snapshot_data is not None:
            first_snapshot_data["snapshot_type"] = SnapshotType.temporary
            snapshots.insert(0, first_snapshot_data)
        return snapshots



from __future__ import annotations

from collections.abc import MutableMapping
from dataclasses import asdict, dataclass, field
from datetime import datetime as dt
from typing import Any, Self

import pytest

from timelineomat import BaseTMSnapshot, SnapshotType, extract_snapshot_data

pytestmark = pytest.mark.anyio

obj1 = object()
obj2 = object()
obj3 = object()
obj4 = object()


async def asyncify(inp):
    return inp


@dataclass(kw_only=True)
class DummySnapshot(BaseTMSnapshot):
    snapshot_for: dt = field(init=False)
    content_specifier: str = field(init=False)
    data: dict[str, Any] = field(init=False)
    snapshot_type: SnapshotType = field(init=False)

    stay_same: Any
    stringified: str
    stringified_int: int

    @classmethod
    def deserialize_snapshot(
        cls,
        full_data: MutableMapping[str, Any],
        *,
        data: MutableMapping[str, Any],
        snapshot_for: dt,
        snapshot_type: SnapshotType,
        content_specifier: str,
    ) -> Self:
        full_data["stringified_int"] = int(full_data["stringified_int"])
        full_data["stringified"] = str(full_data["stringified"])
        full_data.pop("content_specifier", None)
        full_data.pop("snapshot_for", None)
        return super().deserialize_snapshot(
            full_data,
            data=data,
            snapshot_for=snapshot_for,
            snapshot_type=snapshot_type,
            content_specifier=content_specifier,
        )

    def save(self):
        full_data = asdict(self)
        full_data["stringified_int"] = str(full_data["stringified_int"])
        for key in list(full_data.keys()):
            if key not in self.updated_attrs:
                del full_data[key]
        return full_data

    @classmethod
    def get_snapshots_impl(cls, *, content_specifier, after, before, start_snapshot, shall_async=False):
        snapshots = [start_snapshot] if start_snapshot is not None else []
        first_snapshot_data = None
        for snap in globals()[content_specifier]:
            extracted, timepoint, snap_type = extract_snapshot_data(snap, fallback_tz=cls.snapshot_fallback_tz)
            if before is not None and timepoint >= before:
                break
            if after is not None and timepoint < after:
                if snap_type == SnapshotType.full:
                    first_snapshot_data = extracted
                    first_snapshot_data["snapshot_for"] = timepoint
                elif first_snapshot_data is not None:
                    first_snapshot_data.update(extracted)
                    first_snapshot_data["snapshot_for"] = timepoint
                continue
            if snap_type != SnapshotType.full and not snapshots and first_snapshot_data is None:
                raise ValueError("No full snapshot found")
            if not snapshots and first_snapshot_data is not None:
                if snap_type != SnapshotType.full:
                    first_snapshot_data["snapshot_type"] = SnapshotType.temporary
                    first_snapshot_data["snapshot_for"] = after or timepoint
                    assert "snapshot_type" in first_snapshot_data
                    snapshots.append(first_snapshot_data)

                first_snapshot_data = None
            # start from the last full snapshot if both are None
            if snap_type == SnapshotType.full and before is None and after is None:
                snapshots.clear()
                first_snapshot_data = None
            assert "snapshot_type" in snap
            snapshots.append(snap)
        if first_snapshot_data is not None:
            first_snapshot_data["snapshot_type"] = SnapshotType.temporary
            snapshots.insert(0, first_snapshot_data)
        if shall_async:
            return asyncify(snapshots)
        return snapshots


@dataclass(kw_only=True)
class DummySnapshot2(DummySnapshot):
    @classmethod
    def get_snapshots(cls, **kwargs):
        return super().get_snapshots(**kwargs)

    @classmethod
    def process_snapshot_content_specifier(cls, content_specifier: str | None):
        return content_specifier or "dummy_snapshots"


dummy_snapshots = [
    {
        "snapshot_for": dt(year=2025, month=1, day=1),
        "snapshot_type": SnapshotType.full,
        "content_specifier": "DjangoContentType",
        "stay_same": obj1,
        "stringified": 1,
        "stringified_int": 8,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=2),
        "snapshot_type": SnapshotType.sparse,
        "content_specifier": "DjangoContentType",
        "stringified_int": 8,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=3),
        "snapshot_type": SnapshotType.sparse,
        "content_specifier": "DjangoContentType",
        "stringified": 9,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=4),
        "snapshot_type": SnapshotType.full,
        "content_specifier": "DjangoContentType",
        "stay_same": obj2,
        "stringified": 111,
        "stringified_int": 10,
    },
]

dummy_snapshots2 = [
    *dummy_snapshots,
    {
        "snapshot_for": dt(year=2025, month=1, day=5),
        "snapshot_type": SnapshotType.sparse,
        "content_specifier": "DjangoContentType",
        "stringified": 2,
    },
]

dummy_snapshots_empty: list[Any] = []


def test_invalid_invocation():
    with pytest.raises(TypeError):
        DummySnapshot.get_snapshots(",", "foo")
    with pytest.raises(TypeError):
        DummySnapshot.get_snapshots("foo", content_specifier="foo")


def test_invalid_start_snapshot():
    with pytest.raises(ValueError):
        snapshot = DummySnapshot(
            stay_same=obj1,
            stringified=1,
            stringified_int=8,
        )
        snapshot.data = {}
        snapshot.snapshot_for = dt(year=2025, month=1, day=1)
        snapshot.snapshot_type = SnapshotType.sparse
        snapshot.content_specifier = "DjangoContentType"
        DummySnapshot.get_snapshots("dummy_snapshots", start_snapshot=snapshot)

    with pytest.raises(ValueError):
        DummySnapshot.get_snapshots(
            "dummy_snapshots",
            start_snapshot={
                "snapshot_for": dt(year=2025, month=1, day=1),
                "snapshot_type": SnapshotType.sparse,
                "content_specifier": "DjangoContentType",
                "stay_same": obj1,
                "stringified": 1,
                "stringified_int": 8,
            },
        )


def test_overwrite():
    assert isinstance(next(iter(DummySnapshot2.get_snapshots()))[1], DummySnapshot2)
    assert isinstance(next(iter(DummySnapshot2.get_snapshots(content_specifier="dummy_snapshots")))[1], DummySnapshot2)
    timeline = DummySnapshot2.get_snapshots(content_specifier="dummy_snapshots_empty")
    assert timeline[dt(year=2025, month=1, day=1)] is None
    assert list(timeline) == []


def test_empty():
    assert list(DummySnapshot.get_snapshots("dummy_snapshots_empty")) == []


def test_snapshot_unranged():
    timeline = DummySnapshot.get_snapshots("dummy_snapshots")
    assert len(timeline) == 1
    snap = timeline[None]
    assert snap.stringified_int == 10
    assert snap.stringified == "111"
    assert snap.stay_same is obj2
    assert snap.snapshot_type == SnapshotType.full
    assert snap.snapshot_for == dt(year=2025, month=1, day=4)


async def test_async_arg():
    timeline = await DummySnapshot.get_snapshots("dummy_snapshots", shall_async=True)
    snap = timeline[None]
    assert len(timeline) == 1
    assert snap.stringified_int == 10
    assert snap.stringified == "111"
    assert snap.stay_same is obj2
    assert snap.snapshot_type == SnapshotType.full
    assert snap.snapshot_for == dt(year=2025, month=1, day=4)


def test_snapshot_before():
    snap = DummySnapshot.get_snapshots("dummy_snapshots", before=dt(year=2025, month=1, day=4))[None]
    assert snap.stringified_int == 8
    assert snap.stringified == "9"
    assert snap.stay_same is obj1
    assert snap.snapshot_type == SnapshotType.sparse


def test_snapshot_before_all():
    snap = DummySnapshot.get_snapshots("dummy_snapshots", before=dt(year=2025, month=1, day=1))[None]
    assert snap is None


def test_snapshot_after_all_full():
    timeline = DummySnapshot.get_snapshots("dummy_snapshots", after=dt(year=2025, month=1, day=10))
    assert len(timeline) == 1
    snap = timeline[None]
    assert snap is not None
    assert snap.snapshot_for == dt(year=2025, month=1, day=4)
    assert snap.snapshot_type == SnapshotType.temporary


def test_snapshot_after_all_sparse():
    timeline = DummySnapshot.get_snapshots("dummy_snapshots2", after=dt(year=2025, month=1, day=10))
    assert len(timeline) == 1
    snap = timeline[None]
    assert snap is not None
    assert snap.snapshot_for == dt(year=2025, month=1, day=5)
    assert snap.snapshot_type == SnapshotType.temporary



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Awaitable, Generator, Iterator, MutableMapping, Sequence
from datetime import datetime as dt
from datetime import timedelta as td
from enum import IntEnum
from functools import cached_property
from inspect import isawaitable
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Literal,
    NoReturn,
    Protocol,
    Self,
    TypedDict,
    TypeVar,
    cast,
    overload,
)


class SnapshotType(IntEnum):
    sparse = 0
    full = 1
    temporary = 2


SnapshotTypeVariant = SnapshotType | Literal["sparse", "full", "temporary"]


class SnapshotObject(Protocol):
    data: MutableMapping[str, Any] | dict[str, Any]
    snapshot_for: dt | int | str
    snapshot_type: SnapshotTypeVariant


class SnapshotDict(TypedDict):
    data: MutableMapping[str, Any] | dict[str, Any]
    snapshot_for: dt | int | str
    snapshot_type: SnapshotType


SnapshotDict.__required_keys__ = frozenset(("snapshot_for", "snapshot_type"))


SnapshotTimelineEntry = SnapshotObject | SnapshotDict


def parse_dt(inp: dt | int | str, *, fallback_tz: Any = None) -> dt:
    if isinstance(inp, str):
        inp = dt.fromisoformat(inp)
    elif isinstance(inp, int):
        inp = dt.fromtimestamp(inp)
    if inp.tzinfo is None and fallback_tz is not None:
        inp = inp.replace(tzinfo=fallback_tz)
    return inp


def extract_snapshot_data(
    entry: SnapshotTimelineEntry, *, fallback_tz: Any = None
) -> tuple[dict[str, Any], dt, SnapshotType]:
    if isinstance(entry, dict):
        # data can be flat or in data when a dict
        if "data" in entry:
            data: dict[str, Any] = {**entry["data"]}
            snapshot_type: SnapshotTypeVariant = entry["snapshot_type"]
            snapshot_for = entry["snapshot_for"]
        else:
            data = {**entry}
            snapshot_type = data.pop("snapshot_type")
            snapshot_for = data.pop("snapshot_for")
    else:
        # data must be in "data" attribute
        data = {**entry.data}
        snapshot_for = entry.snapshot_for
        snapshot_type = entry.snapshot_type

    # int or stringified int
    if isinstance(snapshot_type, str):
        if snapshot_type in SnapshotType.__members__:
            snapshot_type = cast(SnapshotType, getattr(SnapshotType, snapshot_type))
        elif snapshot_type.isdecimal():
            snapshot_type = SnapshotType(int(snapshot_type))
        else:
            raise ValueError("Invalid snapshot_type.")
    elif isinstance(snapshot_type, int):
        snapshot_type = SnapshotType(snapshot_type)
    elif not isinstance(snapshot_type, SnapshotType):
        raise ValueError("Invalid snapshot_type.")
    return (data, parse_dt(snapshot_for, fallback_tz=fallback_tz), snapshot_type)


_SnapshotInstanceType = TypeVar("_SnapshotInstanceType", bound="BaseTMSnapshot")


class TMSnapshotTimeline(Generic[_SnapshotInstanceType]):
    _snapshots: list[_SnapshotInstanceType]
    klass: type[_SnapshotInstanceType]

    def __init__(
        self, klass: type[_SnapshotInstanceType], snapshots: Sequence[SnapshotTimelineEntry], content_specifier: str, /
    ) -> None:
        self.klass = klass
        self._snapshots = []
        data: dict[str, Any] | None = None
        for snapshot in snapshots:
            extracted, timepoint, snapshot_type = extract_snapshot_data(
                snapshot, fallback_tz=klass.snapshot_fallback_tz
            )
            if snapshot_type == SnapshotType.full or snapshot_type == SnapshotType.temporary:
                data = extracted
            elif data is not None:
                data.update(extracted)
            else:
                raise ValueError("Must start with a full or temporary snapshot")
            assert isinstance(data, MutableMapping)
            self._snapshots.append(
                klass.deserialize_snapshot(
                    # manipulation will lead to hard to debug issues otherwise
                    data.copy(),
                    data=extracted,
                    snapshot_for=timepoint,
                    snapshot_type=snapshot_type,
                    content_specifier=content_specifier,
                )
            )

    @classmethod
    async def resolve_async(
        cls,
        klass: type[_SnapshotInstanceType],
        snapshots: Awaitable[Sequence[SnapshotTimelineEntry]],
        content_specifier: str,
    ) -> Self:
        sn = await snapshots
        return cls(klass, sn, content_specifier)

    def __getitem__(self, timestamp: dt | int | str | None, /) -> _SnapshotInstanceType | None:
        if timestamp is None:
            if self._snapshots:
                return self._snapshots[-1]
            return None
        sanitized_timestamp: dt = parse_dt(timestamp, fallback_tz=self.klass.snapshot_fallback_tz)
        last_item: _SnapshotInstanceType | None = None
        for item in self._snapshots:
            item_sanitized_timestamp = parse_dt(item.snapshot_for, fallback_tz=self.klass.snapshot_fallback_tz)
            if item_sanitized_timestamp > sanitized_timestamp:
                # use last_item, current item is not applicable anymore
                return last_item
            last_item = item
        return last_item

    def iterate(
        self, *, step: td | None = None, before: dt | None = None, after: dt | None = None
    ) -> Generator[tuple[dt, _SnapshotInstanceType], None, None]:
        counter: dt = cast(dt, None)
        pos_item: int = 0
        next_timestamp: None | dt = None
        len_snapshots: int = len(self._snapshots)
        if not len_snapshots:
            return
        # test before ensures len_snapshots > 0
        max_pos_snapshots: int = len_snapshots - 1
        while True:
            item = self._snapshots[pos_item]
            if counter is None or step is None:
                # step is None, implies counter is of snapshot
                counter = parse_dt(item.snapshot_for, fallback_tz=self.klass.snapshot_fallback_tz)
            if before is not None and counter >= before:
                return
            if after is None or counter > after:
                yield (counter, item)
            if step is not None:
                counter += step
                if pos_item < max_pos_snapshots:
                    if next_timestamp is None:
                        next_timestamp = parse_dt(item.snapshot_for, fallback_tz=self.klass.snapshot_fallback_tz)
                    if next_timestamp <= counter:
                        pos_item += 1
                        next_timestamp = None
            else:
                if pos_item < max_pos_snapshots:
                    pos_item += 1
                else:
                    return

    def __iter__(self) -> Iterator[tuple[dt, _SnapshotInstanceType]]:
        return self.iterate()

    def __len__(self):
        return len(self._snapshots)

    def __bool__(self):
        return bool(self._snapshots)


_SnapshotImplReturnType = TypeVar("_SnapshotImplReturnType", covariant=True)
_SnapshotImplKwargs = TypeVar("_SnapshotImplKwargs")


class _BaseSnapshotImplType(Protocol[_SnapshotImplReturnType]):
    @classmethod
    @abstractmethod
    def get_snapshots_impl(
        cls,
        *,
        content_specifier: str,
        after: dt | None,
        before: dt | None,
        start_snapshot: None | SnapshotTimelineEntry,
    ) -> _SnapshotImplReturnType:
        pass


_SyncSnapshotImplType = _BaseSnapshotImplType[Sequence[_SnapshotInstanceType]]
_AsyncSnapshotImplType = _BaseSnapshotImplType[Awaitable[Sequence[_SnapshotInstanceType]]]

if TYPE_CHECKING:

    class _BaseTMSnapshot(_BaseSnapshotImplType, SnapshotObject, ABC):
        content_specifier: str

else:
    # otherwise we end with an inconsistent inheritance order e.g. with pydantic

    class _BaseTMSnapshot:
        # don't use protocol nor predefine the attrs

        @classmethod
        def get_snapshots_impl(
            cls,
            *,
            content_specifier: str,
            after: dt | None,
            before: dt | None,
            start_snapshot: None | SnapshotTimelineEntry,
        ) -> NoReturn:
            # instead of abc or protocol and altering the metaclass, we just raise an error
            # this approach works also with pydantic
            raise NotImplementedError('Error: "get_snapshots_impl" must be implemented.')


class BaseTMSnapshot(_BaseTMSnapshot):
    snapshot_fallback_tz: ClassVar[Any] = None
    snapshot_attrs_as_kwargs: ClassVar[bool] = False

    @classmethod
    def deserialize_snapshot(
        cls,
        full_data: MutableMapping[str, Any],
        *,
        data: MutableMapping[str, Any],
        snapshot_for: dt,
        snapshot_type: SnapshotType,
        content_specifier: str,
    ) -> Self:
        snapshot_obj = cls(**full_data)
        snapshot_obj.data = data
        snapshot_obj.snapshot_type = snapshot_type
        snapshot_obj.snapshot_for = snapshot_for
        snapshot_obj.content_specifier = content_specifier
        return snapshot_obj

    @classmethod
    def process_snapshot_content_specifier(cls, content_specifier: str | None) -> str:
        """
        For customizing the naming logic and processing, like escaping.

        The result is used for comparations.

        By default None and "" are handled equally but this logic is overwritable.
        """
        return content_specifier or cls.__name__

    @overload
    @classmethod
    def get_snapshots(
        cls: type[_AsyncSnapshotImplType],
        content_specifier: str | None = None,
        *,
        after: dt | None = None,
        before: dt | None = None,
        start_snapshot: None | SnapshotTimelineEntry = None,
        **kwargs: _SnapshotImplKwargs,
    ) -> Awaitable[TMSnapshotTimeline[_SnapshotInstanceType]]: ...

    @overload
    @classmethod
    def get_snapshots(
        cls: type[_SyncSnapshotImplType],
        content_specifier: str | None = None,
        *,
        after: dt | None = None,
        before: dt | None = None,
        start_snapshot: None | SnapshotTimelineEntry = None,
        **kwargs: _SnapshotImplKwargs,
    ) -> TMSnapshotTimeline[_SnapshotInstanceType]: ...

    @classmethod
    def get_snapshots(
        cls: type[_AsyncSnapshotImplType | _SyncSnapshotImplType],
        content_specifier: str | None = None,
        *,
        after: dt | None = None,
        before: dt | None = None,
        start_snapshot: None | SnapshotTimelineEntry = None,
        **kwargs: _SnapshotImplKwargs,
    ) -> TMSnapshotTimeline[_SnapshotInstanceType] | Awaitable[TMSnapshotTimeline[_SnapshotInstanceType]]:
        content_specifier = cast(BaseTMSnapshot, cls).process_snapshot_content_specifier(content_specifier)
        if start_snapshot is not None:
            snap_for, snap_type = extract_snapshot_data(start_snapshot)[1:]
            if snap_type not in {
                SnapshotType.full,
                SnapshotType.temporary,
            }:
                raise ValueError("Invalid `start_snapshot`. Not a `temporary` or `full` snapshot.")
            # the start snapshot ensures that after is set
            if after is None or after < snap_for:
                after = snap_for
            if isinstance(start_snapshot, dict):
                content_specifier_start = start_snapshot.get("content_specifier")
            else:
                content_specifier_start = getattr(start_snapshot, "content_specifier", None)
            # if found compare against the model type of start_snapshot
            if content_specifier_start:
                assert content_specifier_start == content_specifier, (
                    "`start_snapshot` and `content_specifier` doesn't match."
                )
        result = cls.get_snapshots_impl(
            content_specifier=content_specifier, after=after, before=before, start_snapshot=start_snapshot, **kwargs
        )
        if isawaitable(result):
            return TMSnapshotTimeline.resolve_async(cast(type[_SnapshotInstanceType], cls), result, content_specifier)
        else:
            return TMSnapshotTimeline(cast(type[_SnapshotInstanceType], cls), result, content_specifier)

    @cached_property
    def updated_attrs(self) -> set[str]:
        """
        Return set with updated attrs.

        Use `self.__dict__.pop("updated_attrs", None)` to reset.
        """
        return set()

    @property
    def object_not_updated(self) -> bool:
        """Object was not touched and no relevant attributes were changed."""
        return "updated_attrs" not in self.__dict__

    def is_attr_change_update(self, name: str, value: Any) -> bool:
        """Update to limit to actual fields."""
        if "data" not in self.__dict__:
            # we are in init
            return False
        return name not in self.data or self.data[name] != value

    def __setattr__(self, name: str, value: Any):
        if self.is_attr_change_update(name, value):
            self.updated_attrs.add(name)
        elif "updated_attrs" in self.__dict__:
            self.updated_attrs.discard(name)
        return super().__setattr__(name, value)

    def __delattr__(self, name: str):
        if name in self.data:
            self.updated_attrs.add(name)
        elif "updated_attrs" in self.__dict__:
            self.updated_attrs.discard(name)
        return super().__delattr__(name)



from collections.abc import MutableMapping
from datetime import datetime as dt
from typing import Any, Self

import edgy
import pytest
from pydantic import BaseModel, Field, computed_field

from timelineomat import BaseTMSnapshot, SnapshotType, extract_snapshot_data

DB_URL = ...
models = edgy.Registry(database=DB_URL)

pytestmark = pytest.mark.anyio


class SnapshotImplementation(edgy.Model):
    data = edgy.fields.JSONField(default=dict)
    snapshot_for: dt = edgy.fields.DateTimeField()
    snapshot_type = edgy.fields.SmallIntegerField()
    # if you want to validate that no temporary snapshot_type is materialized
    # snapshot_type = edgy.fields.SmallIntegerField(gte=0, lt=2)
    content_specifier = edgy.fields.CharField(max_length=10)

    class Meta:
        registry = models


# note: not an edgy model
class Snapshot(BaseTMSnapshot, BaseModel):
    data: MutableMapping[str, Any] = Field(init=False, default_factory=dict)
    snapshot_for: dt
    snapshot_type: SnapshotType
    _loaded: bool = False

    @classmethod
    def deserialize_snapshot(
        cls,
        full_data: MutableMapping[str, Any],
        *,
        data: MutableMapping[str, Any],
        snapshot_for: dt,
        snapshot_type: SnapshotType,
        content_specifier: str,
    ) -> Self:
        snapshot_obj = cls(**full_data)
        snapshot_obj.data = data
        snapshot_obj.snapshot_type = snapshot_type
        snapshot_obj.snapshot_for = snapshot_for
        # snapshot_obj.content_specifier = content_specifier
        snapshot_obj._loaded = True
        return snapshot_obj

    @classmethod
    async def get_snapshots_impl(cls, *, content_specifier, after=None, before=None, start_snapshot=None):
        snapshots = [start_snapshot] if start_snapshot is not None else []
        first_snapshot_data = None
        query = SnapshotImplementation.query.filter(content_specifier=content_specifier).order_by("snapshot_for")

        if before is not None:
            query = query.filter(snapshot_for__lt=before)
        if after is None:
            # start_snapshot is None
            assert start_snapshot is None
            starting_snapshot = await query.filter(snapshot_type=SnapshotType.full).last()
            if starting_snapshot is not None:
                snapshots.append(starting_snapshot)
                after = extract_snapshot_data(starting_snapshot)[1]
                query = query.filter(snapshot_for__gt=after)
            snapshots_list = await query
        elif start_snapshot is not None:
            query = query.filter(snapshot_for__gt=after)
            snapshots_list = await query
        else:
            starting_snapshot = await query.filter(snapshot_type=SnapshotType.full, snapshot_for__lt=after).last()
            if starting_snapshot is not None:
                query = query.filter(snapshot_for__gt=starting_snapshot.snapshot_for)
                snapshots_list = [starting_snapshot, *(await query)]
            else:
                snapshots_list = await query
            # otherwise just load everything and filter manually
        for snap in snapshots_list:
            extracted, timepoint, snap_type = extract_snapshot_data(snap, fallback_tz=cls.snapshot_fallback_tz)
            if after is not None and timepoint < after:
                if snap_type == SnapshotType.full:
                    first_snapshot_data = extracted
                    first_snapshot_data["snapshot_for"] = timepoint
                elif first_snapshot_data is not None:
                    first_snapshot_data.update(extracted)
                    first_snapshot_data["snapshot_for"] = timepoint
                continue
            if snap_type != SnapshotType.full and not snapshots and first_snapshot_data is None:
                raise ValueError("No full snapshot found")
            if not snapshots and first_snapshot_data is not None:
                if snap_type != SnapshotType.full:
                    first_snapshot_data["snapshot_type"] = SnapshotType.temporary
                    first_snapshot_data["snapshot_for"] = after or timepoint
                    snapshots.append(first_snapshot_data)

                first_snapshot_data = None
            snapshots.append(snap)
        if first_snapshot_data is not None:
            first_snapshot_data["snapshot_type"] = SnapshotType.temporary
            snapshots.insert(0, first_snapshot_data)
        return snapshots

    @classmethod
    def get_snapshots(cls, **kwargs):
        assert kwargs.get("content_specifier") is None
        return super().get_snapshots(**kwargs)

    async def save(self):
        snapshot_for, snap_type = extract_snapshot_data(self, fallback_tz=self.snapshot_fallback_tz)[1:]
        if snap_type == SnapshotType.temporary:
            raise ValueError("Cannot save a temporary Snapshot.")
        if self._loaded and not self.object_not_updated:
            return
        data = self.model_dump(include=self.updated_attrs if self._loaded else None, exclude_none=True)
        impl = await SnapshotImplementation(
            data=data, snapshot_for=snapshot_for, snapshot_type=snap_type, content_specifier=type(self).__name__
        ).save()
        self.data = data
        self.__dict__.pop("updated_attrs", None)
        self._loaded = True
        return impl


class SnapshotSubtype1(Snapshot):
    string_field: str
    int_field: int

    @computed_field
    def content_specifier(self) -> str:
        return type(self).__name__


class SnapshotForIdObjects(Snapshot):
    id: int
    other_type: Any

    @computed_field
    def content_specifier(self) -> str:
        return f"{type(self).__name__}:{self.id}"



from .snapshots import (
    BaseTMSnapshot,
    SnapshotType,
    SnapshotTypeVariant,
    TMSnapshotTimeline,
    extract_snapshot_data,
)
from .timeline import (
    NoCallAllowedError,
    PositionsOffsetsTuple,
    SkipEmptyEvent,
    SkipEvent,
    SkipInvalidEvent,
    SkipOccludedEvent,
    TimelineOMat,
    TimeRangeTuple,
    ordered_insert,
    streamline_event,
    streamline_event_times,
    streamlined_ordered_insert,
    transform_events_to_times,
)

__version__ = "1.0.0"

__all__ = [
    "streamline_event_times",
    "streamline_event",
    "streamlined_ordered_insert",
    "transform_events_to_times",
    "ordered_insert",
    "TimelineOMat",
    "TMSnapshotTimeline",
    "SkipEvent",
    "SkipInvalidEvent",
    "SkipEmptyEvent",
    "SkipOccludedEvent",
    "NoCallAllowedError",
    "PositionsOffsetsTuple",
    "TimeRangeTuple",
    "BaseTMSnapshot",
    "SnapshotType",
    "SnapshotTypeVariant",
    "extract_snapshot_data",
]



from collections.abc import MutableMapping
from datetime import datetime as dt
from datetime import timedelta as td
from typing import Annotated, Any, Self

import edgy
import pytest
from edgy import Registry
from edgy.testing.client import DatabaseTestClient
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, computed_field

from timelineomat import BaseTMSnapshot, SnapshotType, extract_snapshot_data

database = DatabaseTestClient("sqlite:///test_db.sqlite", drop_database=True)
models = Registry(database=edgy.Database(database, force_rollback=True))

pytestmark = pytest.mark.anyio


@pytest.fixture(autouse=True, scope="module")
async def create_test_database():
    # this creates and drops the database
    async with database:
        await models.create_all()
        yield
        await models.drop_all()


@pytest.fixture(autouse=True, scope="function")
async def rollback_transactions():
    # this rolls back
    async with models:
        yield


class SnapshotImplementation(edgy.Model):
    data = edgy.fields.JSONField(default=dict)
    snapshot_for: dt = edgy.fields.DateTimeField(primary_key=True)
    content_specifier = edgy.fields.CharField(max_length=50, primary_key=True)
    # only 0, 1
    snapshot_type = edgy.fields.SmallIntegerField(gte=0, lt=2)

    class Meta:
        registry = models


# Note: not an edgy model
class Snapshot(BaseTMSnapshot, BaseModel):
    data: MutableMapping[str, Any] = Field(init=False, default_factory=dict)
    snapshot_for: dt
    snapshot_type: SnapshotType
    _loaded: bool = False

    @classmethod
    def deserialize_snapshot(
        cls,
        full_data: MutableMapping[str, Any],
        *,
        data: MutableMapping[str, Any],
        snapshot_for: dt,
        snapshot_type: SnapshotType,
        content_specifier: str,
    ) -> Self:
        snapshot_obj = cls(**full_data)
        snapshot_obj.data = data
        snapshot_obj.snapshot_type = snapshot_type
        snapshot_obj.snapshot_for = snapshot_for
        # snapshot_obj.content_specifier = content_specifier
        snapshot_obj._loaded = True
        return snapshot_obj

    @classmethod
    async def get_snapshots_impl(cls, *, content_specifier, after=None, before=None, start_snapshot=None):
        snapshots = [start_snapshot] if start_snapshot is not None else []
        first_snapshot_data = None
        query = SnapshotImplementation.query.filter(content_specifier=content_specifier).order_by("snapshot_for")

        if before is not None:
            query = query.filter(snapshot_for__lt=before)
        if after is None:
            # start_snapshot is None
            assert start_snapshot is None
            starting_snapshot = await query.filter(snapshot_type=SnapshotType.full).last()
            if starting_snapshot is not None:
                snapshots.append(starting_snapshot)
                after = extract_snapshot_data(starting_snapshot)[1]
                query = query.filter(snapshot_for__gt=after)
            snapshots_list = await query
        elif start_snapshot is not None:
            query = query.filter(snapshot_for__gt=after)
            snapshots_list = await query
        else:
            starting_snapshot = await query.filter(snapshot_type=SnapshotType.full, snapshot_for__lt=after).last()
            if starting_snapshot is not None:
                query = query.filter(snapshot_for__gt=starting_snapshot.snapshot_for)
                snapshots_list = [starting_snapshot, *(await query)]
            else:
                snapshots_list = await query
            # otherwise just load everything and filter manually
        for snap in snapshots_list:
            extracted, timepoint, snap_type = extract_snapshot_data(snap, fallback_tz=cls.snapshot_fallback_tz)
            if after is not None and timepoint < after:
                if snap_type == SnapshotType.full:
                    first_snapshot_data = extracted
                    first_snapshot_data["snapshot_for"] = timepoint
                elif first_snapshot_data is not None:
                    first_snapshot_data.update(extracted)
                    first_snapshot_data["snapshot_for"] = timepoint
                continue
            if snap_type != SnapshotType.full and not snapshots and first_snapshot_data is None:
                raise ValueError("No full snapshot found")
            if not snapshots and first_snapshot_data is not None:
                if snap_type != SnapshotType.full:
                    first_snapshot_data["snapshot_type"] = SnapshotType.temporary
                    first_snapshot_data["snapshot_for"] = after or timepoint
                    snapshots.append(first_snapshot_data)

                first_snapshot_data = None
            snapshots.append(snap)
        if first_snapshot_data is not None:
            first_snapshot_data["snapshot_type"] = SnapshotType.temporary
            snapshots.insert(0, first_snapshot_data)
        return snapshots

    @classmethod
    def get_snapshots(cls, **kwargs):
        assert kwargs.get("content_specifier") is None
        # content_specifier = cls.__name__
        return super().get_snapshots(**kwargs)

    async def save(self):
        snapshot_for, snap_type = extract_snapshot_data(self, fallback_tz=self.snapshot_fallback_tz)[1:]
        if snap_type == SnapshotType.temporary:
            raise ValueError("Cannot save a temporary Snapshot.")
        if self._loaded and not self.object_not_updated:
            return
        data = self.model_dump(include=self.updated_attrs if self._loaded else None, exclude_none=True)
        impl = await SnapshotImplementation(
            data=data, snapshot_for=snapshot_for, snapshot_type=snap_type, content_specifier=type(self).__name__
        ).save()
        self.data = data
        self.__dict__.pop("updated_attrs", None)
        self._loaded = True
        return impl


def stringify(x):
    return str(x)


class SnapshotSubtype1(Snapshot):
    stringified: Annotated[str | None, BeforeValidator(stringify)] = None
    stringified_int: Annotated[int | None, PlainSerializer(stringify)] = None

    @computed_field
    def content_specifier(self) -> str:
        return type(self).__name__

    @classmethod
    def deserialize_snapshot(
        cls,
        full_data: MutableMapping[str, Any],
        *,
        data: MutableMapping[str, Any],
        snapshot_for: dt,
        snapshot_type: SnapshotType,
        content_specifier: str,
    ) -> Self:
        if full_data.get("stringified_int") is not None:
            full_data["stringified_int"] = int(full_data["stringified_int"])
        if full_data.get("stringified") is not None:
            full_data["stringified"] = str(full_data["stringified"])
        return super().deserialize_snapshot(
            full_data,
            data=data,
            snapshot_for=snapshot_for,
            snapshot_type=snapshot_type,
            content_specifier=content_specifier,
        )


class SnapshotForIdObjects(Snapshot):
    id: int
    stringified2: Annotated[str | None, BeforeValidator(stringify)] = None
    stringified_int2: Annotated[int | None, PlainSerializer(stringify)] = None

    @computed_field
    def content_specifier(self) -> str:
        return f"{type(self).__name__}:{self.id}"

    @classmethod
    def deserialize_snapshot(
        cls,
        full_data: MutableMapping[str, Any],
        *,
        data: MutableMapping[str, Any],
        snapshot_for: dt,
        snapshot_type: SnapshotType,
        content_specifier: str,
    ) -> Self:
        if full_data.get("stringified_int2") is not None:
            full_data["stringified_int2"] = int(full_data["stringified_int2"])
        if full_data.get("stringified2") is not None:
            full_data["stringified2"] = str(full_data["stringified2"])
        return super().deserialize_snapshot(
            full_data,
            data=data,
            snapshot_for=snapshot_for,
            snapshot_type=snapshot_type,
            content_specifier=content_specifier,
        )


async def test_violations():
    with pytest.raises(ValueError):
        await SnapshotImplementation.query.create(
            snapshot_type=30, snapshot_for=dt(year=2025, month=1, day=2), content_specifier="foo"
        )
    with pytest.raises(ValueError):
        await SnapshotImplementation.query.create(
            snapshot_type=SnapshotType.full, snapshot_for=dt(year=2025, month=1, day=2)
        )


async def test_bad_invovation():
    with pytest.raises(TypeError):
        SnapshotSubtype1.get_snapshots("foo")
    with pytest.raises(AssertionError):
        SnapshotSubtype1.get_snapshots(content_specifier="foo")


sample_snapshots_sub1 = [
    {
        "snapshot_for": dt(year=2025, month=1, day=2),
        "snapshot_type": SnapshotType.full,
        "stringified": 1,
        "stringified_int": 7,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=2, hour=2),
        "snapshot_type": SnapshotType.sparse,
        "stringified_int": 8,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=3),
        "snapshot_type": SnapshotType.sparse,
        "stringified": 9,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=4),
        "snapshot_type": SnapshotType.full,
        "stringified": 111,
        "stringified_int": 10,
    },
    {
        "snapshot_for": dt(year=2025, month=1, day=5),
        "snapshot_type": SnapshotType.sparse,
        "stringified": 11,
    },
]


async def test_basic():
    for i in sample_snapshots_sub1:
        impl = await SnapshotSubtype1.model_validate(i, strict=False).save()
        assert impl.data
    tl = await SnapshotForIdObjects.get_snapshots()
    assert len(tl) == 0
    real = await SnapshotSubtype1.get_snapshots()
    assert len(real) == 2
    assert real[dt(year=2025, month=1, day=4)].snapshot_type == SnapshotType.full
    last_snap = real[dt(year=2025, month=1, day=5)]
    assert last_snap.snapshot_type == SnapshotType.sparse
    assert last_snap.snapshot_for == dt(year=2025, month=1, day=5)
    assert last_snap.stringified == "11"
    assert last_snap.stringified_int == 10
    after_last_snap = real[dt(year=2025, month=1, day=8)]
    assert after_last_snap.snapshot_type == SnapshotType.sparse
    assert after_last_snap.snapshot_for == dt(year=2025, month=1, day=5)
    assert after_last_snap.stringified == "11"
    assert after_last_snap.stringified_int == 10

    combined = await SnapshotSubtype1.get_snapshots(before=dt(year=2025, month=1, day=4))
    assert len(combined) == 3
    assert combined[dt(year=2025, month=1, day=3)] == combined[None]
    assert combined[dt(year=2025, month=1, day=3)].snapshot_type == SnapshotType.sparse
    assert combined[dt(year=2025, month=1, day=3)].stringified == "9"


async def test_get_all():
    for i in sample_snapshots_sub1:
        await SnapshotSubtype1.model_validate(i, strict=False).save()

    tl = await SnapshotSubtype1.get_snapshots(after=dt.min)
    assert len(tl) == len(sample_snapshots_sub1)
    assert len(list(tl.iterate(before=dt(year=2025, month=1, day=4)))) == 3
    assert len(list(tl.iterate(after=dt(year=2025, month=1, day=4)))) == 1


async def test_iterate_step():
    for i in sample_snapshots_sub1:
        await SnapshotSubtype1.model_validate(i, strict=False).save()

    tl = await SnapshotSubtype1.get_snapshots()
    assert (
        len(
            list(tl.iterate(after=dt(year=2025, month=1, day=4), before=dt(year=2025, month=1, day=7), step=td(days=1)))
        )
        == 2
    )



import contextlib
from dataclasses import dataclass
from datetime import UTC
from datetime import datetime as dt
from datetime import timedelta as td

import pytest
from faker import Faker

import timelineomat


@dataclass
class Event1:
    start: dt
    stop: dt


@dataclass
class Event2:
    begin: dt
    end: dt


def _generate_time_tuple(faker, start):
    return start, start + td(hours=faker.random_int(1, 48))


def _generate_event_series(_type, _variant):
    faker = Faker()
    ts_start = faker.past_datetime(
        "-200d",
    )
    events = []
    for _i in range(1000):
        ts_stop = _generate_time_tuple(faker, ts_start)[1]
        if _variant == 1:
            events.append(_type(start=ts_start, stop=ts_stop))
        else:
            events.append(_type(begin=ts_start, end=ts_stop))
        ts_start += td(hours=faker.random_int(1, 48))
    return events


def test_event1_direct():
    events = _generate_event_series(Event1, 1)
    events_finished = []
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(timelineomat.streamline_event(ev, events_finished))
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event.stop <= ev.start
        last_event = ev


def test_event2_direct():
    events = _generate_event_series(Event2, 2)
    events_finished = []
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(
                timelineomat.streamline_event(
                    ev,
                    events_finished,
                    start_extractor="begin",
                    stop_extractor="end",
                )
            )
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event.end <= ev.begin
        last_event = ev


def test_dict1_direct():
    events = _generate_event_series(dict, 1)
    events_finished = []
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(
                timelineomat.streamline_event(
                    ev,
                    events_finished,
                )
            )
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event["start"] <= ev["stop"]
        last_event = ev


def test_dict2_direct():
    events = _generate_event_series(dict, 2)
    events_finished = []
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(
                timelineomat.streamline_event(
                    ev,
                    events_finished,
                    start_extractor="begin",
                    stop_extractor="end",
                )
            )
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event["end"] <= ev["begin"]
        last_event = ev


def test_event1_timelineomat():
    events = _generate_event_series(Event1, 1)
    events_finished = []
    tm = timelineomat.TimelineOMat()
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(tm.streamline_event(ev, events_finished))
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event.stop <= ev.start
        last_event = ev


def test_event2_timelineomat():
    events = _generate_event_series(Event2, 2)
    events_finished = []
    tm = timelineomat.TimelineOMat(start_extractor="begin", stop_extractor="end")
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(tm.streamline_event(ev, events_finished))
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event.end <= ev.begin
        last_event = ev


def test_dict1_timelineomat():
    events = _generate_event_series(dict, 1)
    tm = timelineomat.TimelineOMat()
    events_finished = []
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(
                tm.streamline_event(
                    ev,
                    events_finished,
                )
            )
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event["start"] <= ev["stop"]
        last_event = ev


def test_dict2_timelineomat():
    events = _generate_event_series(dict, 2)
    tm = timelineomat.TimelineOMat(start_extractor="begin", stop_extractor="end")
    events_finished = []
    for ev in events:
        with contextlib.suppress(timelineomat.SkipEvent):
            events_finished.append(
                tm.streamline_event(
                    ev,
                    events_finished,
                )
            )
    last_event = None
    for ev in events_finished:
        if last_event:
            assert last_event["end"] <= ev["begin"]
        last_event = ev


def test_event_empty():
    events = []
    with pytest.raises(timelineomat.SkipEmptyEvent):
        timelineomat.streamline_event_times(
            Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 2)), events, ensure_timespan=True
        )
    timelineomat.streamline_event_times(
        Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 2)), events, ensure_timespan=False
    )


def test_event_empty_collision_raises():
    events = [Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 2))]
    with pytest.raises(timelineomat.SkipOccludedEvent):
        timelineomat.streamline_event_times(Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 2)), events)


def test_event_getting_empty_raises():
    events = [Event1(start=dt(2024, 2, 1), stop=dt(2024, 3, 2)), Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 3))]
    with pytest.raises(timelineomat.SkipOccludedEvent):
        timelineomat.streamline_event_times(Event1(start=dt(2024, 3, 1), stop=dt(2024, 3, 2, 1)), events)


def test_invalid():
    events = []
    with pytest.raises(timelineomat.SkipInvalidEvent):
        timelineomat.streamline_event_times(Event1(start=dt(2024, 3, 2), stop=dt(2024, 2, 1)), events)


def test_event_within_event():
    events = [Event1(start=dt(2024, 1, 1), stop=dt(2024, 3, 1))]
    new_event = Event1(start=dt(2024, 1, 2), stop=dt(2024, 2, 1))
    occlusions = []
    with pytest.raises(timelineomat.SkipEvent):
        timelineomat.streamline_event_times(new_event, events, occlusions=occlusions)
    assert new_event.start == occlusions[0].start
    assert new_event.stop == occlusions[0].stop


def test_handle_timestamps():
    events = [
        Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 2)),
        Event1(start=dt(2024, 3, 4), stop=dt(2024, 3, 4)),
        Event1(start=dt(2024, 3, 4), stop=dt(2024, 3, 8)),
    ]
    ev = Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 5))
    # fake insert
    result = timelineomat.streamlined_ordered_insert(ev, events, direction="desc", no_update_timelines={0})
    assert ev.stop == dt(2024, 3, 4)
    assert result.positions == [1]
    assert result.offsets == [2]
    assert len(events) == 3
    # let it raise (cut skip)
    with pytest.raises(timelineomat.SkipEvent):
        timelineomat.streamlined_ordered_insert(Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 5)), events)
    # insert it for real
    timelineomat.ordered_insert(ev, events)
    assert events[result.positions[0]] is ev


def test_handle_timestamp_cut_handle():
    events = [
        Event1(start=dt(2024, 3, 1), stop=dt(2024, 3, 1)),
        Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 2)),
        Event1(start=dt(2024, 3, 4), stop=dt(2024, 3, 4)),
        Event1(start=dt(2024, 3, 7), stop=dt(2024, 3, 7)),
        Event1(start=dt(2024, 3, 8), stop=dt(2024, 3, 8)),
    ]
    ev = Event1(start=dt(2024, 3, 2), stop=dt(2024, 3, 5))

    def cut_handler(new_tup, orig_tup, remaining):
        remaining = list(remaining)
        assert len(remaining) == 3
        assert remaining[0] == events[2]
        # cut on remaining[0]
        return (new_tup.start, remaining[0].start)

    tm = timelineomat.TimelineOMat(direction="desc", cut_handler=cut_handler)
    result = tm.streamlined_ordered_insert(ev, events)
    assert ev.stop == dt(2024, 3, 4)
    assert result.positions == [2]
    assert result.offsets == [3]


def test_result():
    timeline = [Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)), Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3))]
    new_event = Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 4))
    # one time methods
    occlusions = []
    assert timelineomat.streamline_event_times(
        new_event, timeline, occlusions=occlusions
    ) == timelineomat.TimeRangeTuple(start=dt(2024, 1, 3), stop=dt(2024, 1, 4))
    assert occlusions[0] == timelineomat.TimeRangeTuple(start=dt(2024, 1, 1), stop=dt(2024, 1, 3))
    # TimelineOMat method
    tm = timelineomat.TimelineOMat()
    occlusions = []
    assert tm.streamline_event_times(new_event, timeline, occlusions=occlusions) == timelineomat.TimeRangeTuple(
        start=dt(2024, 1, 3), stop=dt(2024, 1, 4)
    )


def test_result_fallback_utc():
    timeline = [Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)), Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3))]
    new_event = Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 4))
    # one time function
    assert timelineomat.streamline_event_times(
        new_event, timeline, fallback_timezone=UTC
    ) == timelineomat.TimeRangeTuple(start=dt(2024, 1, 3, tzinfo=UTC), stop=dt(2024, 1, 4, tzinfo=UTC))


def one_time_overwrite_end(ev):
    if isinstance(ev, dict):
        return ev["end"]
    else:
        return ev.stop


def test_onetime_overwrite():
    timeline = [Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)), Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3))]
    new_event1 = Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 4))
    new_event2 = dict(start=dt(2024, 1, 1).timestamp(), end=dt(2024, 1, 5).timestamp())

    tm = timelineomat.TimelineOMat()
    timeline.append(tm.streamline_event(new_event1, timeline))
    assert timeline[-1].stop == dt(2024, 1, 4)
    assert timeline[-1].start == dt(2024, 1, 3)
    timeline.append(
        Event1(**tm.streamline_event_times(new_event2, timeline, stop_extractor=one_time_overwrite_end)._asdict())
    )
    assert timeline[-1].stop == dt(2024, 1, 5)
    assert timeline[-1].start == dt(2024, 1, 4)
    # test conversion in TimeRangeTuple array
    assert [t for t, ev in tm.transform_events_to_times(timeline)] == [
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)),
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 2), stop=dt(2024, 1, 3)),
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 3), stop=dt(2024, 1, 4)),
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 4), stop=dt(2024, 1, 5)),
    ]
    # test sorting

    assert [
        t for t, ev in tm.transform_events_to_times(sorted(timeline, key=tm.streamline_event_times, reverse=True))
    ] == [
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 4), stop=dt(2024, 1, 5)),
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 3), stop=dt(2024, 1, 4)),
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 2), stop=dt(2024, 1, 3)),
        timelineomat.TimeRangeTuple(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)),
    ]


@pytest.mark.parametrize("direction", ["asc", "desc"])
def test_ordered_insert_multiline(direction):
    timeline1 = [
        Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)),
        # invalid event
        {},
        Event1(start=dt(2024, 1, 5), stop=dt(2024, 1, 6)),
        Event1(start=dt(2024, 1, 10), stop=dt(2024, 1, 11)),
        Event1(start=dt(2024, 1, 12), stop=dt(2024, 1, 13)),
    ]
    timeline2 = []
    timeline3_init = (Event1(start=dt(2024, 1, 11), stop=dt(2024, 1, 12)),)
    timeline3 = tuple(timeline3_init)
    tm = timelineomat.TimelineOMat(direction=direction, no_update_timelines={2})
    position_offset = tm.ordered_insert(
        Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3)),
        timeline1,
        timeline2,
        timeline3,
        direction=direction,
    )
    # invalid element is skipped
    if direction == "asc":
        assert position_offset == ([2, 0, 0], [2, 0, 0])
    else:
        assert position_offset == ([1, 0, 0], [4, 0, 1])
        # unset it for desc, we add asc events
        position_offset[1][0] = 0
        position_offset[1][1] = 0
    assert timeline3 == timeline3_init
    # test stability
    position_offset = tm.ordered_insert(
        Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3)),
        timeline1,
        timeline2,
        timeline3,
        offsets=position_offset.offsets,
    )
    if direction == "asc":
        assert position_offset == ([3, 1, 0], [3, 1, 0])
    else:
        assert position_offset == ([1, 0, 0], [5, 1, 1])
        position_offset[1][0] = 0
        position_offset[1][1] = 0
    assert timeline3 == timeline3_init
    # overlapping
    position_offset = tm.ordered_insert(
        Event1(start=dt(2024, 1, 7), stop=dt(2024, 1, 12)),
        timeline1,
        timeline2,
        timeline3,
        offset=position_offset,
    )
    if direction == "asc":
        assert position_offset == ([5, 2, 0], [5, 2, 0])
    else:
        # pass empty elements
        assert position_offset == ([6, 1, 0], [1, 1, 1])
        position_offset[1][0] = 0


def test_ordered_insert_desc():
    # desc is more complicated
    timeline = [
        Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)),
        # invalid event
        {},
        Event1(start=dt(2024, 1, 5), stop=dt(2024, 1, 6)),
        Event1(start=dt(2024, 1, 10), stop=dt(2024, 1, 11)),
        Event1(start=dt(2024, 1, 12), stop=dt(2024, 1, 13)),
    ]
    tm = timelineomat.TimelineOMat(direction="desc")
    # we need to insert descending
    positions, offsets = tm.ordered_insert(Event1(start=dt(2024, 1, 12), stop=dt(2024, 1, 13)), timeline)
    assert offsets == [1]
    assert positions == [4]
    positions, offsets = tm.ordered_insert(
        Event1(start=dt(2024, 1, 12), stop=dt(2024, 1, 13)), timeline, offsets=offsets
    )
    assert offsets == [2]
    positions, offsets = tm.ordered_insert(Event1(start=dt(2024, 1, 7), stop=dt(2024, 1, 8)), timeline, offsets=offsets)
    assert offsets == [4]
    positions = tm.ordered_insert(
        Event1(start=dt(2023, 1, 12), stop=dt(2023, 1, 13)), timeline, offsets=offsets
    ).positions
    assert positions == [0]
    # we are still descendend concerning the 2nd last insert and didn't updated the offset
    positions = tm.ordered_insert(
        Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3)), timeline, offsets=offsets
    ).positions
    assert positions == [2]


def test_streamlined_ordered_insert_desc():
    # desc is more complicated
    timeline = [
        Event1(start=dt(2024, 1, 1), stop=dt(2024, 1, 2)),
        # invalid event
        {},
        Event1(start=dt(2024, 1, 5), stop=dt(2024, 1, 6)),
        Event1(start=dt(2024, 1, 10), stop=dt(2024, 1, 11)),
        Event1(start=dt(2024, 1, 12), stop=dt(2024, 1, 13)),
    ]
    tm = timelineomat.TimelineOMat(direction="desc")
    # we need to insert descending
    with pytest.raises(timelineomat.SkipEvent):
        tm.streamlined_ordered_insert(Event1(start=dt(2024, 1, 12), stop=dt(2024, 1, 13)), timeline)
    positions, offsets = tm.streamlined_ordered_insert(Event1(start=dt(2024, 1, 7), stop=dt(2024, 1, 11)), timeline)
    assert offsets == [2]
    assert timeline[positions[0]].start == dt(2024, 1, 7)
    assert timeline[positions[0]].stop == dt(2024, 1, 10)
    position = tm.streamlined_ordered_insert(
        Event1(start=dt(2023, 1, 12), stop=dt(2023, 1, 13)), timeline, offsets=offsets
    ).positions[0]
    assert position == 0
    # we are still descendend concerning the 2nd last insert and didn't updated the offset
    position = tm.streamlined_ordered_insert(
        Event1(start=dt(2024, 1, 2), stop=dt(2024, 1, 3)), timeline, offsets=offsets
    ).positions[0]
    assert position == 2


def test_invalid_rejection():
    timeline = []
    new_event1 = Event1(stop=dt(2024, 1, 1), start=dt(2024, 1, 4))
    with pytest.raises(timelineomat.SkipEvent):
        timelineomat.streamline_event_times(new_event1, timeline)
    with pytest.raises(timelineomat.SkipEvent):
        timelineomat.streamline_event_times(new_event1)
    with pytest.raises(timelineomat.SkipEvent):
        timelineomat.ordered_insert(new_event1, timeline)

    tm = timelineomat.TimelineOMat()
    with pytest.raises(timelineomat.SkipEvent):
        tm.streamline_event_times(new_event1, timeline)
    with pytest.raises(timelineomat.SkipEvent):
        tm.streamline_event_times(new_event1)
    with pytest.raises(timelineomat.SkipEvent):
        tm.ordered_insert(new_event1, timeline)
